
package sga;

/**
 * This main class represents population and implements evolution process.
 * @author fanda
 */
public class SGA {

    /**
     * Fixed length of chromosomes
     */
    public static int maxGenes = 24;				
    
    /**
     * Minimum and maximum values of variable x
     */
    public static double minX = -5.0;				
    public static double maxX = 5.0;				

    /**
     * Population size
     */
    public static int popSize = 100;				
    
    /**
     * Stopping criterion
     */
    public static int evaluations = 10000;		
    
    /**
     * Crossover type 0...uniform, 1...1-point, 2...2-point
     */
    public static int crossType = 1;				
    
    /**
     * Selection type 0...tournament selection, 1...roulette wheel selection
     */
    public static int selectionType = 1;
    
    /**
     * Parameters of tournament selection
     */
    public static int tournamentSize = 3;
    
    /**
     * Probability of crossover
     */
    public static double pC = 0.75;					
    
    /**
     * Probability of mutation
     */
    public static double pM = 0.5;						
        
    /**
     * Parents population.
     */
    static Individual[] parentPop;
    
    /**
     * Old population.
     */
    static Individual[] oldPop;
    
    /**
     * Newly generated individuals.
     */
    static Individual[] offspringPop;    

    /**
     * SGA parameters
     */
    public static SGAParameters parameters = SGAParameters.getInstance();
    
    public static void initParameters(){
        
        popSize = parameters.getPopSize();
        evaluations = parameters.getEvaluations();
        crossType = parameters.getCrossType();
        pC = parameters.getPC();
        pM = parameters.getPM();
        selectionType = parameters.getSelectionType();
        tournamentSize = parameters.getTournamentSize();
    }
    
    /**
     * Main function.
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception{
        
        parameters.readConfigFile();
        initParameters();
       
        int cgen = 0;
        int totEvals = 0;
        
        // Allocate memory for arrays.
        parentPop = new Individual[popSize];
        oldPop = new Individual[popSize];
        offspringPop = new Individual[2];
        
        // Initialise population.
        popInit();
        
        // Generate statistic data.
        StatData.getStatistics(oldPop);
        // Print results.
        StatData.reportResults(totEvals);
        StatData.printBest(oldPop);
        
        // While ending criterion is not satisfied...
        while(totEvals < evaluations) {
            int chroms_generated = 0;
            cgen++;
            
            // Elitism. Copy best individual from old population.
            copyIndividual(StatData.indbest, oldPop, chroms_generated++, parentPop);
            
            // While new population is not filled...
            while(chroms_generated < popSize) {
                int[] par = selection(oldPop);
                int par1 = par[0];
                int par2 = par[1];
                // Creates two new individuals.
                if(Math.random() < pC) {
                    offspringPop = oldPop[par1].crossover(oldPop[par2]);
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                else {
                    copyIndividual(par1, oldPop, 0, offspringPop);
                    copyIndividual(par2, oldPop, 1, offspringPop);
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                // Calculates fitness of new individuals.
                offspringPop[0].get_fitness();
                offspringPop[1].get_fitness();
                totEvals += 2;  // Increase number of evaluations done.
                
                // Get offspring to new parent population.
                copyIndividual(0, offspringPop, chroms_generated++, parentPop);
                if(chroms_generated < popSize) 
                    copyIndividual(1, offspringPop, chroms_generated++, parentPop);
                
            }
            
            // Get parent population to old population and repeat while cycle.
            Individual[] tempPop = new Individual[popSize];
            for(int i = 0; i < popSize; i++) {
                tempPop[i] = oldPop[i].clone();
                oldPop[i] = parentPop[i].clone();
                parentPop[i] = tempPop[i].clone();
            }
            
            // Generate statistic data.
            StatData.getStatistics(oldPop);
            // Print results.
            StatData.reportResults(totEvals);
            StatData.printBest(oldPop);
        
        } 
    }
    
    /**
     * Initialization of population.
     */
    public static void popInit() {
        // Allocate memory for population.
        parentPop = new Individual[popSize];
        offspringPop = new Individual[popSize];
        
        // Generate individuals in parent population and calculate their fitness values.
        for(int ind = 0; ind < popSize; ind++) {
            oldPop[ind] = new Individual();
            oldPop[ind].get_fitness(); // Evaluate fitness of new individual.
            //old_pop[ind] = parent_pop[ind].clone();
        } 
        
    }
    
    /**
     * Makes deep copy of individual in population p with index from and stores it to population p2 to index to.
     * @param from index of original individual
     * @param p population with original individual
     * @param to index of copied individual in new population 
     * @param p2 new population
     */
    public static void copyIndividual(int from, Individual[] p, int to, Individual[] p2) {
        p2[to] = p[from].clone();
    }
    
    /**
     * Tournament selection with number of individuals in tournament.
     * @return index of winner individual.
     */
    public static int tournamentSelection(Individual[] p) {
        int index;
        int winnerIndex = (int)(Math.random() * (popSize)); // First member of tournament selection.
        double winnerFitness = p[winnerIndex].fitness;

        // Choose N individuals and select best one.
        // We chose randomly first individual 3 rows up, so there is only SGA.tournament_size-1 random individuals to check. 
        for(int individualIndex = 0; individualIndex < tournamentSize-1; individualIndex++) {
            index = (int)(Math.random() * (popSize));
            if(p[index].fitness > winnerFitness) {
                winnerFitness = p[index].fitness;
                winnerIndex = index;
            }
        }
        
        return winnerIndex;
    }
    
    /**
     * Roulette wheel selection with totalForestFitness weight.
     * @param p population which from are individuals chosen.
     * @param totalForestFitness sum of fitness values of all individuals in population p.
     * @return index of selected individual.
     */
    public static int rouletteWheelSelection(Individual[] p) {
        int winnerIndex = -1;

        double rnd = Math.random() * (StatData.fsum + popSize*Math.abs(StatData.fworst));
        double sum = 0;
        
        // Choose individual with probability according fitness sum of population.
        for(int i = 0; i < popSize; i++) {
            //sum += p[StatData.indworst].fitness - p[i].fitness;   // If wee minimise fitnes value.
            sum += p[i].fitness + Math.abs(StatData.fworst);    // If we maximise fitness value.
            if(sum > rnd) {
                winnerIndex = i;
                break;
            }
        }
            
        if(winnerIndex == -1) {
            winnerIndex = popSize-1;
        }  
        
        return winnerIndex;      
    }    
    
    /**
     * Performs selection method according selectionType value.
     * @return array of two indexes selected by selection methods.
     */
    public static int[] selection(Individual[] p) {
        int[] chosen = new int[2];
        
        switch(selectionType) {
            case 0: 
                chosen[0] = tournamentSelection(p);
                chosen[1] = chosen[0];
                while(chosen[0] == chosen[1]) 
                    chosen[1] = tournamentSelection(p);
                break;
            case 1:
                chosen[0] = rouletteWheelSelection(p);
                chosen[1] = chosen[0];
                while(chosen[0] == chosen[1]) 
                    chosen[1] = rouletteWheelSelection(p);
                break;
            default:
                System.err.println("ERROR - Bad value of selectionType variable in SGA.java file. Check if its value is 0 or 1.");
                System.exit(1);
        }
        
        return chosen;
    }
}
